@@ -9,7 +9,10 @@ import warnings
 import pandas as pd
 import plotly.io as pio
 import itertools
+import functools
+import requests
 
+from requests.exceptions import RequestException
 from bravado_core.param import Param
 from bson.objectid import ObjectId
 from typing import Union, Type, List
@@ -19,7 +22,7 @@ from pathlib import Path
 from copy import deepcopy
 from filetype import guess
 from flatten_dict import flatten, unflatten
-from base64 import b64encode, b64decode
+from base64 import b64encode, b64decode, urlsafe_b64encode
 from urllib.parse import urlparse
 from pyisemail import is_email
 from collections import defaultdict
@@ -74,6 +77,9 @@ VALID_URLS |= {f"http://localhost.{n}-api.materialsproject.org" for n in SUBDOMA
 SUPPORTED_FILETYPES = (Gz, Jpeg, Png, Gif, Tiff)
 SUPPORTED_MIMES = [t().mime for t in SUPPORTED_FILETYPES]
 DEFAULT_DOWNLOAD_DIR = Path.home() / "mpcontribs-downloads"
+EMPTY_SPEC_DICT = {
+    "swagger": "2.0", "paths": {}, "info": {"title": "Swagger", "version": "0.0"},
+}
 
 j2h = Json2Html()
 pd.options.plotting.backend = "plotly"
@@ -335,12 +341,6 @@ class Attachment(dict):
         )
 
 
-def load_client(apikey=None, headers=None, host=None):
-    warnings.warn(
-        "load_client(...) is deprecated, use Client(...) instead", DeprecationWarning
-    )
-
-
 def _run_futures(futures, total: int = 0, timeout: int = -1, desc=None):
     """helper to run futures/requests"""
     start = time.perf_counter()
@@ -367,12 +367,109 @@ def _run_futures(futures, total: int = 0, timeout: int = -1, desc=None):
                 timed_out = timeout > 0 and elapsed > timeout
 
                 if timed_out or not response.ok:
-                    for future in futures:
-                        future.cancel()
+                    for fut in futures:
+                        fut.cancel()
 
     return responses
 
 
+@functools.lru_cache
+def _load(protocol, host, headers_json, project):
+    headers = ujson.loads(headers_json)
+    http_client = FidoClientGlobalHeaders(headers=headers)
+    url = f"{protocol}://{host}"
+    origin_url = f"{url}/apispec.json"
+    apispec = Path(urlsafe_b64encode(origin_url.encode('utf-8')).decode('utf-8'))
+
+    if apispec.exists():
+        spec_dict = ujson.loads(apispec.read_bytes())
+        print(f"Specs for {origin_url} re-loaded from {apispec}.")
+    else:
+        try:
+            if requests.options(f"{url}/healthcheck").status_code == 200:
+                loader = Loader(http_client)
+                spec_dict = loader.load_spec(origin_url)
+
+                with apispec.open("w") as f:
+                    ujson.dump(spec_dict, f)
+
+                print(f"Specs for {origin_url} saved as {apispec}.")
+            else:
+                spec_dict = EMPTY_SPEC_DICT
+                print(f"Specs not loaded: Healthcheck for {url} failed!")
+        except RequestException:
+            spec_dict = EMPTY_SPEC_DICT
+            print(f"Specs not loaded: Could not connect to {url}!")
+
+    spec_dict["host"] = host
+    spec_dict["schemes"] = [protocol]
+
+    config = {
+        "validate_responses": False,
+        "use_models": False,
+        "include_missing_properties": False,
+        "formats": [email_format, url_format],
+    }
+    bravado_config = bravado_config_from_config_dict(config)
+    for key in set(bravado_config._fields).intersection(set(config)):
+        del config[key]
+    config["bravado"] = bravado_config
+    swagger_spec = Spec.from_dict(spec_dict, origin_url, http_client, config)
+
+    if not spec_dict["paths"]:
+        return swagger_spec
+
+    # expand regex-based query parameters for `data` columns
+    session = get_session()
+    query = {"name": project} if project else {}
+    query["_fields"] = ["columns"]
+    kwargs = dict(headers=headers, params=query)
+    future = session.get(f"{url}/projects/", **kwargs)
+    track_id = "get_columns"
+    setattr(future, "track_id", track_id)
+    resp = _run_futures([future], timeout=3).get(track_id, {}).get("result")
+    session.close()
+
+    if not resp or not resp["data"]:
+        return swagger_spec
+
+    if project and not resp["data"]:
+        raise ValueError(f"{project} doesn't exist, or access denied!")
+
+    columns = {"string": [], "number": []}
+
+    for proj in resp["data"]:
+        for column in proj["columns"]:
+            if column["path"].startswith("data."):
+                col = column["path"].replace(".", "__")
+                if column["unit"] == "NaN":
+                    columns["string"].append(col)
+                else:
+                    col = f"{col}__value"
+                    columns["number"].append(col)
+
+    resource = swagger_spec.resources["contributions"]
+
+    for operation_id, operation in resource.operations.items():
+        for pn in list(operation.params.keys()):
+            if pn.startswith("data_"):
+                param = operation.params.pop(pn)
+                op = param.name.rsplit('$__', 1)[-1]
+                typ = param.param_spec.get("type")
+                key = "number" if typ == "number" else "string"
+
+                for column in columns[key]:
+                    param_name = f"{column}__{op}"
+                    param_spec = deepcopy(param.param_spec)
+                    param_spec["name"] = param_name
+                    param_spec.pop("description", None)
+                    operation.params[param_name] = Param(
+                        swagger_spec, operation, param_spec
+                    )
+
+    return swagger_spec
+
+
 class Client(SwaggerClient):
     """client to connect to MPContribs API
 
@@ -382,13 +479,8 @@ class Client(SwaggerClient):
           >>> from mpcontribs.client import Client
           >>> client = Client()
     """
-    # We only want to load the swagger spec from the remote server when needed and not
-    # everytime the client is initialized. Hence using the Borg design nonpattern (instead
-    # of Singleton): Since the __dict__ of any instance can be re-bound, Borg rebinds it
-    # in its __init__ to a class-attribute dictionary. Now, any reference or binding of an
-    # instance attribute will actually affect all instances equally.
+    # Borg: https://www.oreilly.com/library/view/python-cookbook/0596001673/ch05s23.html
     # NOTE bravado future doesn't work with concurrent.futures
-
     _shared_state = {}
 
     def __init__(
@@ -424,24 +516,20 @@ class Client(SwaggerClient):
         self.headers = headers or {}
         self.headers = {"x-api-key": apikey} if apikey else self.headers
         self.headers["Content-Type"] = "application/json"
+        self.headers_json = ujson.dumps(self.headers, sort_keys=True)
         self.host = host
         ssl = host.endswith(".materialsproject.org") and not host.startswith("localhost.")
         self.protocol = "https" if ssl else "http"
         self.url = f"{self.protocol}://{self.host}"
-        self.session = get_session()
+        self.project = project
 
         if self.url not in VALID_URLS:
             raise ValueError(f"{self.url} not a valid URL (one of {VALID_URLS})")
 
-        if "swagger_spec" not in self.__dict__ or (
-            self.swagger_spec.http_client.headers != self.headers
-        ) or (
-            self.swagger_spec.spec_dict["host"] != self.host
-        ) or (
-            "project" not in self.__dict__ or self.project != project
-        ):
-            self.project = project
-            self._load()
+        if "session" not in self.__dict__:
+            self.session = get_session()
+
+        super().__init__(self.cached_swagger_spec)
 
     def __enter__(self):
         return self
@@ -449,71 +537,9 @@ class Client(SwaggerClient):
     def __exit__(self, exc_type, exc_val, exc_tb):
         return self.session.close()
 
-    def _load(self):
-        http_client = FidoClientGlobalHeaders(headers=self.headers)
-        loader = Loader(http_client)
-        origin_url = f"{self.url}/apispec.json"
-        spec_dict = loader.load_spec(origin_url)
-        spec_dict["host"] = self.host
-        spec_dict["schemes"] = [self.protocol]
-
-        config = {
-            "validate_responses": False,
-            "use_models": False,
-            "include_missing_properties": False,
-            "formats": [email_format, url_format],
-        }
-        bravado_config = bravado_config_from_config_dict(config)
-        for key in set(bravado_config._fields).intersection(set(config)):
-            del config[key]
-        config["bravado"] = bravado_config
-
-        swagger_spec = Spec.from_dict(spec_dict, origin_url, http_client, config)
-        super().__init__(
-            swagger_spec, also_return_response=bravado_config.also_return_response
-        )
-
-        # expand regex-based query parameters for `data` columns
-        try:
-            query = {"name": self.project} if self.project else {}
-            resp = self.projects.get_entries(_fields=["columns"], **query).result()
-        except AttributeError:
-            # skip in tests
-            return
-
-        if self.project and not resp["data"]:
-            raise ValueError(f"{self.project} doesn't exist, or access denied!")
-
-        columns = {"string": [], "number": []}
-
-        for project in resp["data"]:
-            for column in project["columns"]:
-                if column["path"].startswith("data."):
-                    col = column["path"].replace(".", "__")
-                    if column["unit"] == "NaN":
-                        columns["string"].append(col)
-                    else:
-                        col = f"{col}__value"
-                        columns["number"].append(col)
-
-        resource = self.swagger_spec.resources["contributions"]
-
-        for operation_id, operation in resource.operations.items():
-            for pn in list(operation.params.keys()):
-                if pn.startswith("data_"):
-                    param = operation.params.pop(pn)
-                    op = param.name.rsplit('$__', 1)[-1]
-                    typ = param.param_spec.get("type")
-                    key = "number" if typ == "number" else "string"
-
-                    for column in columns[key]:
-                        param_name = f"{column}__{op}"
-                        param_spec = deepcopy(param.param_spec)
-                        param_spec["name"] = param_name
-                        param_spec.pop("description", None)
-                        operation.params[param_name] = Param(
-                            self.swagger_spec, operation, param_spec
-                        )
+    @property
+    def cached_swagger_spec(self):
+        return _load(self.protocol, self.host, self.headers_json, self.project)
 
     def __dir__(self):
         members = set(self.swagger_spec.resources.keys())
@@ -521,6 +547,10 @@ class Client(SwaggerClient):
         members |= set(k for k in dir(self.__class__) if not k.startswith("_"))
         return members
 
+    def _reinit(self):
+        _load.cache_clear()
+        super().__init__(self.cached_swagger_spec)
+
     def _is_valid_payload(self, model: str, data: dict):
         model_spec = deepcopy(self.get_model(f"{model}sSchema")._model_spec)
         model_spec.pop("required")
@@ -624,6 +654,9 @@ class Client(SwaggerClient):
             name (str): name of the project
         """
         name = self.project or name
+        if not name:
+            return {"error": "initialize client with project or set `name` argument!"}
+
         return Dict(self.projects.get_entry(pk=name, _fields=["_all"]).result())
 
     def get_contribution(self, cid: str) -> Type[Dict]:
@@ -739,18 +772,17 @@ class Client(SwaggerClient):
 
         return Attachment(self.attachments.get_entry(pk=aid, _fields=["_all"]).result())
 
-    def init_columns(self, columns: dict) -> dict:
+    def init_columns(self, columns: dict = None) -> dict:
         """initialize columns for a project to set their order and desired units
 
-        The `columns` field tracks the minima and maxima of each `data` field as
-        contributions are submitted. This function should thus be executed before
-        submitting contributions, or all contributions for a project should be deleted to
-        ensure clean initialization of all columns. If columns are not initialized using
-        this function, `submit_contributions` will respect the order of columns as they
-        are submitted and will auto-determine suitable units based on the first
-        contribution containing a respective data column. `init_columns` can be used at
-        any point to reset the order of columns, though. Previously determined `min/max`
-        values for the affected data fields will be respected.
+        The `columns` field of a project tracks the minima and maxima of each `data` field
+        in its contributions. If columns are not initialized before submission using this
+        function, `submit_contributions` will respect the order of columns as they are
+        submitted and will try to auto-determine suitable units.
+
+        `init_columns` can be used at any point to reset the order of columns. Omitting
+        the `columns` argument will re-initialize columns based on the `data` fields of
+        all submitted contributions.
 
         The `columns` argument is a dictionary which maps the data field names to its
         units. Use `None` to indicate that a field is not a quantity (plain string). The
@@ -773,115 +805,114 @@ class Client(SwaggerClient):
         Args:
             columns (dict): dictionary mapping data column to its unit
         """
-        if not isinstance(columns, dict):
-            return {"error": "`columns` argument must be a dict!"}
-
         if not self.project:
             return {"error": "initialize client with project argument!"}
 
-        existing_columns = set()
-        for k, v in columns.items():
-            if k in COMPONENTS:
-                existing_columns.add(k)
-                continue
+        columns = columns or {}
+        new_columns = []
 
-            nesting = k.count(".")
-            if nesting > 4:
-                return {"error": f"Nesting too deep for {k}"}
+        if columns:
+            # check columns input
+            scanned_columns = set()
 
-            for col in existing_columns:
-                if nesting and col.startswith(k):
-                    return {"error": f"duplicate definition of {k} in {col}!"}
+            for k, v in columns.items():
+                if k in COMPONENTS:
+                    scanned_columns.add(k)
+                    continue
 
-                for n in range(1, nesting+1):
-                    if k.rsplit(".", n)[0] == col:
-                        return {"error": f"Ancestor of {k} already defined in {col}!"}
+                nesting = k.count(".")
+                if nesting > 4:
+                    return {"error": f"Nesting too deep for {k}"}
 
-            is_valid_string = isinstance(v, str) and v.lower() != "nan"
-            if not is_valid_string and v is not None:
-                return {"error": f"Unit '{v}' for {k} invalid (use `None` or a non-NaN string)!"}
+                for col in scanned_columns:
+                    if nesting and col.startswith(k):
+                        return {"error": f"duplicate definition of {k} in {col}!"}
 
-            if v != "" and v is not None and v not in ureg:
-                return {"error": f"Unit '{v}' for {k} invalid!"}
+                    for n in range(1, nesting+1):
+                        if k.rsplit(".", n)[0] == col:
+                            return {"error": f"Ancestor of {k} already defined in {col}!"}
 
-            existing_columns.add(k)
+                is_valid_string = isinstance(v, str) and v.lower() != "nan"
+                if not is_valid_string and v is not None:
+                    return {
+                        "error": f"Unit '{v}' for {k} invalid (use `None` or a non-NaN string)!"
+                    }
 
-        # sort to avoid "overlapping columns" error in handsontable's NestedHeaders
-        sorted_columns = flatten(unflatten(columns, splitter="dot"), reducer="dot")
+                if v != "" and v is not None and v not in ureg:
+                    return {"error": f"Unit '{v}' for {k} invalid!"}
 
-        # reconcile with existing columns
-        resp = self.projects.get_entry(pk=self.project, _fields=["columns"]).result()
-        existing_columns, new_columns = {}, []
+                scanned_columns.add(k)
 
-        for col in resp["columns"]:
-            path = col.pop("path")
-            existing_columns[path] = col
+            # sort to avoid "overlapping columns" error in handsontable's NestedHeaders
+            sorted_columns = flatten(unflatten(columns, splitter="dot"), reducer="dot")
 
-        for path, unit in sorted_columns.items():
-            if path in COMPONENTS:
-                new_columns.append({"path": path})
-                continue
+            # reconcile with existing columns
+            resp = self.projects.get_entry(pk=self.project, _fields=["columns"]).result()
+            existing_columns = {}
 
-            full_path = f"data.{path}"
-            new_column = {"path": full_path}
-            existing_column = existing_columns.get(full_path)
+            for col in resp["columns"]:
+                path = col.pop("path")
+                existing_columns[path] = col
 
-            if unit is not None:
-                new_column["unit"] = unit
+            for path, unit in sorted_columns.items():
+                if path in COMPONENTS:
+                    new_columns.append({"path": path})
+                    continue
 
-            if existing_column:
-                for k in ["min", "max"]:
-                    v = existing_column.get(k)
-                    if v:
-                        new_column[k] = v
+                full_path = f"data.{path}"
+                new_column = {"path": full_path}
+                existing_column = existing_columns.get(full_path)
 
-                # NOTE if existing_unit == "NaN":
-                #   it was set by omitting "unit" in new_column
-                new_unit = new_column.get("unit", "NaN")
-                existing_unit = existing_column.get("unit")
-                if existing_unit != new_unit:
-                    try:
-                        ureg.Quantity(existing_unit).to(new_unit)
-                    except DimensionalityError:
-                        return {
-                            "error": f"Can't convert {existing_unit} to {new_unit} for {path}"
-                        }
+                if unit is not None:
+                    new_column["unit"] = unit
 
-                    # TODO scale contributions to new unit
-                    return {"error": "Changing units not supported yet. Please resubmit"
-                            " contributions or update accordingly."}
+                if existing_column:
+                    # NOTE if existing_unit == "NaN":
+                    #   it was set by omitting "unit" in new_column
+                    new_unit = new_column.get("unit", "NaN")
+                    existing_unit = existing_column.get("unit")
+                    if existing_unit != new_unit:
+                        try:
+                            ureg.Quantity(existing_unit).to(new_unit)
+                        except DimensionalityError:
+                            return {
+                                "error": f"Can't convert {existing_unit} to {new_unit} for {path}"
+                            }
 
-            new_columns.append(new_column)
+                        # TODO scale contributions to new unit
+                        return {"error": "Changing units not supported yet. Please resubmit"
+                                " contributions or update accordingly."}
+
+                new_columns.append(new_column)
 
         payload = {"columns": new_columns}
         valid = self._is_valid_payload("Project", payload)
         if not valid:
             return {"error": valid}
 
-        self.projects.update_entry(pk=self.project, project={"columns": []}).result()
         return self.projects.update_entry(pk=self.project, project=payload).result()
 
     def delete_contributions(self, query: dict = None, timeout: int = -1):
         """Remove all contributions for a query
 
-        Note: This also resets the columns field for a project. It might have to be
-        re-initialized via `client.init_columns()`.
-
         Args:
             query (dict): optional query to select contributions
             timeout (int): cancel remaining requests if timeout exceeded (in seconds)
         """
-        if not self.project or not query or "project" not in query:
+        if not self.project and (not query or "project" not in query):
             print("initialize client with project, or include project in query!")
             return
 
         tic = time.perf_counter()
         query = query or {}
-        query["project"] = self.project
-        cids = list(self.get_all_ids(query).get(self.project, {}).get("ids", set()))
+
+        if self.project:
+            query["project"] = self.project
+
+        cids = list(self.get_all_ids(query).get(query["project"], {}).get("ids", set()))
 
         if not cids:
-            print(f"There aren't any contributions to delete for {self.project}")
+            print(f"There aren't any contributions to delete for {query['project']}")
             return
 
         total = len(cids)
@@ -892,7 +923,8 @@ class Client(SwaggerClient):
         _run_futures(futures, total=total, timeout=timeout)
         left, _ = self.get_totals(query=query)
         deleted = total - left
-        self._load()
+        self.init_columns()
+        self._reinit()
         toc = time.perf_counter()
         dt = (toc - tic) / 60
         print(f"It took {dt:.1f}min to delete {deleted} contributions.")
@@ -1209,7 +1241,7 @@ class Client(SwaggerClient):
 
         query = query or {}
 
-        if not self.project or not query or "project" not in query:
+        if not self.project and (not query or "project" not in query):
             return {"error": "initialize client with project, or include project in query!"}
 
         if "project" in query and self.project != query["project"]:
@@ -1242,7 +1274,8 @@ class Client(SwaggerClient):
             new_paths = set(c["path"] for c in resp["columns"])
 
             if new_paths != old_paths:
-                self._load()
+                self.init_columns()
+                self._reinit()
 
         toc = time.perf_counter()
         return {"updated": updated, "total": total, "seconds_elapsed": toc - tic}
@@ -1294,22 +1327,27 @@ class Client(SwaggerClient):
             recursive (bool): also set `is_public` for according contributions?
             timeout (int): cancel remaining requests if timeout exceeded (in seconds)
         """
-        if not self.project or not query or "project" not in query:
+        if not self.project and (not query or "project" not in query):
             return {"error": "initialize client with project, or include project in query!"}
 
+        query = query or {}
+
+        if self.project:
+            query["project"] = self.project
+
         try:
-            resp = self.projects.get_entry(pk=self.project, _fields=["is_public"]).result()
+            resp = self.projects.get_entry(pk=query["project"], _fields=["is_public"]).result()
         except HTTPNotFound:
-            return {"error": f"project `{self.project}` not found or access denied!"}
+            return {"error": f"project `{query['project']}` not found or access denied!"}
 
         if not recursive and resp["is_public"] == is_public:
-            return {"warning": f"`is_public` already set to {is_public} for `{self.project}`."}
+            return {"warning": f"`is_public` already set to {is_public} for `{query['project']}`."}
 
         ret = {}
 
         if resp["is_public"] != is_public:
             resp = self.projects.update_entry(
-                pk=self.project, project={"is_public": is_public}
+                pk=query["project"], project={"is_public": is_public}
             ).result()
             ret["published"] = resp["is_public"] == is_public
 
@@ -1613,7 +1651,7 @@ class Client(SwaggerClient):
                     responses = _run_futures(
                         futures, total=ncontribs, timeout=timeout, desc="Submit"
                     )
-                    processed = sum(r["count"] for r in responses.values())
+                    processed = sum(r.get("count", 0) for r in responses.values())
                     total_processed += processed
 
                     if processed != ncontribs and retry and retries < RETRIES and \
@@ -1642,7 +1680,8 @@ class Client(SwaggerClient):
 
             toc = time.perf_counter()
             dt = (toc - tic) / 60
-            self._load()
+            self.init_columns()
+            self._reinit()
             self.session.close()
             self.session = get_session()
             print(f"It took {dt:.1f}min to submit {total_processed}/{total} contributions.")
